# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.AbstractIterableComparator

import logging
logger = logging.getLogger('libracmp')


class DictComparator(libracmp.AbstractIterableComparator.AbstractIterableComparator):
    """dict objects Comparator"""

    @property
    def result(self):
        """comparison property, performs the comparison of dict operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        comparison is done as iteration of iterable items of operands by keys.
        subsequent comparators are used on particular item pairs, -> recursively
        
        if key is missing, None is used as value, then the result depends on next comparator type (mode/filters)
        
        diff is actually computed in result for efficiency.
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        self._diff = {}
        keys = set(self.a_data.keys() + self.b_data.keys())
        self.debug('keys=%s' % (keys,))
        result_vector = []
        for i_key in keys:
            a_item = self.a_data.get(i_key)
            b_item = self.b_data.get(i_key)
            self.debug('i_key=%s, a_item=%s, b_item=%s' % (i_key, a_item, b_item,))
            comparator = self._compare_next(a_item, b_item, i_key)()
            result_vector.append(comparator.result)
            if comparator.diff is not None:
                self._diff[i_key] = comparator.diff

        self._result = all(result_vector)
        if self._result:
            self._diff = None
        self.debug(self)
        return self._result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.AbstractIterableComparator

import logging
logger = logging.getLogger('libracmp')


class Comparator(libracmp.AbstractIterableComparator.AbstractIterableComparator):
    """omni objects Comparator
    does not compare itself, but can be used for various operand types.
    Acts as entry comparator to be used by default.
    """

    @property
    def result(self):
        """comparison property, performs the comparison of dict operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        comparison is done by subsequent specific type comparator.
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        self._diff = None
        comparator = self._compare_next(self.a_data, self.b_data, None)()
        self._result = comparator.result
        if comparator.diff is not None:
            self._diff = comparator.diff

        if self._result:
            self._diff = None
        self.debug(self)
        return self._result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.exceptions
import libracmp.AbstractComparator

import logging
logger = logging.getLogger('libracmp')


class AbstractIterableComparator(libracmp.AbstractComparator.AbstractComparator):
    """Abstract object for Iterable (list) objects Comparator classes in libracmp module"""

    @property
    def result(self):
        """comparison property, performs the comparison of list operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        comparison is done as iteration of iterable items of operands.
        subsequent comparators are used on particular item pairs, -> recursively
        
        if lengths differs, comparison continues, so diff is computed completely.
        
        diff is actually computed in result for efficiency.
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        self._diff = []
        length = max(len(self.a_data), len(self.b_data))
        self.debug('length=%s' % (length,))
        result_vector = []
        for i_indx in xrange(length):
            a_item = None
            b_item = None
            length_error = False
            try:
                a_item = self.a_data[i_indx]
            except IndexError:
                length_error = True
            try:
                b_item = self.b_data[i_indx]
            except IndexError:
                length_error = True
            if not length_error:
                self.debug('i_indx=%s, a_item=%s, b_item=%s' % (i_indx, a_item, b_item,))
                comparator = self._compare_next(a_item, b_item, i_indx)()
                result_vector.append(comparator.result)
                if comparator.diff is not None:
                    self._diff.append(comparator.diff)
            else:
                result_vector.append(False)
                self._diff.append((a_item, b_item))

        self._result = all(result_vector)
        if self._result:
            self._diff = None
        self.debug(self)
        return self._result

    @property
    def diff(self):
        """provides comparison differences if applicable
        diff is computed (by result) and cached, subsequent calls return cached value, object acts as immutable
        :rtype: None, set, list, dict
        :return: comparison differences
        """
        if self._diff is not None:
            return self._diff
        _ = self.result
        return self._diff

    def _compare_next(self, a_next, b_next, i_key):
        """internal subsequent compare of selected item pair
        
        :param a_next: first operand to compare
        :param b_next: second operand to compare
        :type i_key: int, str
        :param i_key: context key / index used for selecting specific filters
        :rtype: lambda
        :return: function that spawn instance of subsequent comparator
        """
        self.debug('mode=%s' % self.mode)
        comparator_next = None
        filters_next = self._select_filters(i_key)

        if (filters_next
                and not isinstance(filters_next, dict)
                and issubclass(filters_next, libracmp.AbstractComparator.AbstractComparator)):
            comparator_next = filters_next
            filters_next = None
        elif isinstance(a_next, dict) and isinstance(b_next, dict):
            comparator_next = self.mode.get('dict')
        elif isinstance(a_next, (list, tuple)) and isinstance(b_next, (list, tuple)):
            comparator_next = self.mode.get('list')

        # getting default
        if comparator_next is None:
            comparator_next = self.mode.get('default')
            if not comparator_next or not issubclass(comparator_next, libracmp.AbstractComparator.AbstractComparator):
                raise libracmp.exceptions.CannotCompareError(
                    'Cannot determine any suitable comparator to use, for current mode \"%s\"' % (self.mode,)
                )
        self.debug('comparator_next=%s' % (comparator_next,))

        return lambda: comparator_next(a_next, b_next, filters_next, self.aux, self.mode, self.depth+1)

    def _select_filters(self, i_key):
        """internal filter selector, used for subsequent comparison
        
        select specific comparator or subfilter, depending on original filter.
        if there is no filter available for specific key, default is selected.
        and if there is no default, None is returned.
        
        If no key is specified, complete filter is returned.
        If no original filter exists Non is returned.
        
        :param i_key: context key, to select filter
        :rtype: dict, libracmp.AbstractIterator.AbstractIterator, None
        :return: selected filter, else default, otherwise None
        """
        if i_key is None or self.filters is None:
            return self.filters
        # select particular filters
        tmp = self.filters.get(i_key)
        if tmp is not None:
            result = tmp
        # select defaults
        else:
            result = self.filters.get(libracmp.All())
        self.debug('_select_filter=%s' % (result,))
        return result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.StrictComparator
import libracmp.RelaxedComparator
import libracmp.TypeStrictComparator
import libracmp.IgnoreComparator
import libracmp.ListComparator
import libracmp.DictComparator

# Basic Strict equality using python ==
# ex: Value comparable values like int(1) == float(1.0) should be equals
MODE_STRICT = {
    'default': libracmp.StrictComparator.StrictComparator,
    'list': libracmp.ListComparator.ListComparator,
    'dict': libracmp.DictComparator.DictComparator,
}

# Basic relaxed equality, acts same as strict, with exception of None
# if one operand is None, then comparison is considered True
# missing Items dict.get() are treated as None and above applies
MODE_RELAXED = {
    'default': libracmp.RelaxedComparator.RelaxedComparator,
    'list': libracmp.ListComparator.ListComparator,
    'dict': libracmp.DictComparator.DictComparator,
}

# Type Strict equality first checks the compatible (equal) types
# ex: int vs floats are always false
MODE_TYPESTRICT = {
    'default': libracmp.TypeStrictComparator.TypeStrictComparator,
    'list': libracmp.ListComparator.ListComparator,
    'dict': libracmp.DictComparator.DictComparator,
}

# Ignore equality treats all operands comparison as True
# useful for filtering specific values
MODE_IGNORE = {
    'default': libracmp.IgnoreComparator.IgnoreComparator,
    'list': libracmp.ListComparator.ListComparator,
    'dict': libracmp.DictComparator.DictComparator,
}

# Default mode used is Basic Strict
MODE_DEFAULT = MODE_STRICT


def singleton(cls):
    """basic singleton decorator"""
    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


@singleton
class All(object):
    """Wildcard key specifier used in filters
    
    Used as object in key, for applying sub-filter or comparator
    for all (default, not defined otherwise) values in such level.
    
    filter={All(): default_comparator, 'mykey': specific_comparator}
    for 'mykey', specific_comparator will be used,
    and for all others, default_comparator will be used. 
    """

    def __str__(self):
        return '<%all>'

    def __repr__(self):
        return str(self)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.ListComparator

import logging
logger = logging.getLogger('libracmp')


class PrefixListComparator(libracmp.ListComparator.ListComparator):
    """List item Comparator"""

    @property
    def result(self):
        """comparison property, performs the length indifferent comparison of list operands 
        from __init__ once computed result is cached, subsequent calls returns 
        cached value, object acts as immutable
        
        comparison is done as iteration of iterable items of operands.
        subsequent comparators are used on particular item pairs, -> recursively
        
        if lengths differs, comparison continues, so diff is computed completely.
        
        diff is actually computed in result for efficiency.
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        self._diff = []
        length = max(len(self.a_data), len(self.b_data))
        self.debug('length=%s' % (length,))
        result_vector = []
        for i_indx in xrange(length):
            a_item = None
            b_item = None
            length_error = False
            try:
                a_item = self.a_data[i_indx]
            except IndexError:
                length_error = True
            try:
                b_item = self.b_data[i_indx]
            except IndexError:
                length_error = True
            if not length_error:
                self.debug('i_indx=%s, a_item=%s, b_item=%s' % (i_indx, a_item, b_item,))
                comparator = self._compare_next(a_item, b_item, i_indx)()
                result_vector.append(comparator.result)
                if comparator.diff is not None:
                    self._diff.append(comparator.diff)
            else:
                break  # Length indifference, comparing others on one side only skipped

        self._result = all(result_vector)
        if self._result:
            self._diff = None
        self.debug(self)
        return self._result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import AbstractComparator

import logging
logger = logging.getLogger('libracmp')


class StringCastComparator(AbstractComparator.AbstractComparator):
    """Item comparator performing string cast before comparison"""

    def __init__(self, a_data, b_data, filters=None, aux=None, mode=None, depth=0):
        """comparison operands are casted to string to perform a string compare
        
        :param a_data: first operand to compare
        :param b_data: second operand to compare
        :type filters: dict
        :param filters: comparison filter
        :param aux: Auxilliary data
        :type mode: dict
        :param mode: Comparison mode for selecting comparators based on types
            dict must contain 3 keys 'default', 'list', 'dict' with selected comparator classes
        :param depth: internal depth modifier
        """
        super(StringCastComparator, self).__init__(a_data, b_data, filters, aux, mode, depth)
        self.a_data = str(a_data)
        self.b_data = str(b_data)



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import libracmp
import logging

logger = logging.getLogger('libracmp')


class AbstractComparator(object):
    """Abstract object for Comparator classes in libracmp module"""

    def __init__(self, a_data, b_data, filters=None, aux=None, mode=None, depth=0):
        """operands are stored in class, to be then compared by .result property use
        
        filters specify exceptional care of specific keys/index (applicable with iterable comparators)
        
        aux is not currently used
        
        mode specifies default selection of comparators for values (applicable with iterable comparators)
        mode has contain 'default', 'list', 'dict' keys with comparator classes to be used for that occasion
        
        depth is internal depth counter used for indented logging, and debugging
        
        :param a_data: first operand to compare
        :param b_data: second operand to compare
        :type filters: dict
        :param filters: comparison filter
        :param aux: Auxilliary data
        :type mode: dict
        :param mode: Comparison mode for selecting comparators based on types
            dict must contain 3 keys 'default', 'list', 'dict' with selected comparator classes
        :param depth: internal depth modifier
        """
        self.a_data = a_data
        self.b_data = b_data
        self.filters = filters
        self.aux = aux
        self.mode = mode or libracmp.MODE_DEFAULT
        self._result = None
        self._diff = None
        self.depth = depth

    @property
    def result(self):
        """comparison property, performs the comparison of operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        self._result = self.a_data == self.b_data
        self.debug(self)
        return self._result

    @property
    def diff(self):
        """provides comparison differences if applicable
        diff is computed and cached, subsequent calls return cached value, object acts as immutable
        :rtype: None, set, list, dict
        :return: comparison differences
        """
        if self._diff is not None:
            return self._diff
        self._diff = None
        if not self.result:
            self._diff = (self.a_data, self.b_data,)
        return self._diff

    def _compatible_types(self, a_item, b_item):
        """determine compatible types
        result might actually depend on used comparator
        
        :param a_item: first comparison operand
        :param b_item: second comparison operand
        :rtype: bool
        :return: type compatibility True if types match
        """
        result = type(a_item) == type(b_item)
        self.debug('a_item=%s, b_item=%s -comp-> %s' % (a_item, b_item, result,))
        return result

    def __str__(self):
        tmp = ("%s(A=%s, B=%s, filters=%s, aux=%s, mode=%s, depth=%s) -> %s diff=%s"
               % (self.__class__.__name__, repr(self.a_data), repr(self.b_data),
                  self.filters, self.aux, self.mode, self.depth, self.result, self.diff))
        return tmp

    def debug(self, msg):
        logger.debug("%s> %s" % (self.depth, msg))



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libracmp.AbstractComparator
import logging
logger = logging.getLogger('libracmp')


class TypeStrictComparator(libracmp.AbstractComparator.AbstractComparator):
    """Item comparator checking for type match before comparison, non-matching types yields False result"""
    @property
    def result(self):
        """comparison property, performs the comparison of operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        first types are checked. Non-matching types yields False results.
        Then regular comparison is performed
        
        :rtype: bool
        :return: comparison result
        """
        if self._result is not None:
            return self._result
        if not self._compatible_types(self.a_data, self.b_data):
            self._result = False
        else:
            self._result = super(TypeStrictComparator, self).result
        return self._result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import AbstractComparator
import logging
logger = logging.getLogger('libracmp')


class RelaxedComparator(AbstractComparator.AbstractComparator):
    """Item comparator when None value is treated as wildcard, yielding True result"""

    @property
    def result(self):
        """comparison property, performs the comparison of operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        any None operand yields True result.
        None is treated as wildcard.
        
        :rtype: bool
        :return: comparison result, with None yielding True
        """
        if self._result is not None:
            return self._result
        if self.a_data is None or self.b_data is None:
            self._result = True
        else:
            self._result = self.a_data == self.b_data
        self.debug(self)
        return self._result

    def _compatible_types(self, a_item, b_item):
        """determine compatible types
        
        any None operand yields True result.
        None is treated as wildcard.
        
        :param a_item: first comparison operand
        :param b_item: second comparison operand
        :rtype: bool
        :return: type compatibility True if types match, with None yielding True
        """
        result = type(a_item) == type(b_item) or (a_item is None) or (b_item is None)
        self.debug('a_item=%s, b_item=%s -comp-> %s' % (a_item, b_item, result,))
        return result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import AbstractComparator
import logging
logger = logging.getLogger('libracmp')


class IgnoreComparator(AbstractComparator.AbstractComparator):
    """Item Comparator for ignoring particular comparison results"""

    @property
    def result(self):
        """comparison property, performs the comparison of operands from __init__
        once computed result is cached, subsequent calls returns cached value, object acts as immutable
        
        Any values are treated as equals.
        
        :rtype: bool
        :return: always returns True
        """
        if self._result is not None:
            return self._result
        self._result = True
        self.debug(self)
        return self._result

    def _compatible_types(self, a_item, b_item):
        """determine compatible types
        
        All values are treated as equal types.
                
        :param a_item: first comparison operand
        :param b_item: second comparison operand
        :rtype: bool
        :return: always returns True
        """
        result = True
        self.debug('a_item=%s, b_item=%s -comp-> %s' % (a_item, b_item, result,))
        return result



# Copyright 2017 Zdenek Kraus <zdenek.kraus@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils.core import setup

setup(
    name='libracmp',
    version='1.0.1',
    packages=['libracmp'],
    url='https://bitbucket.org/zkraus/libracmp',
    license='Apache-2.0',
    author='Zdenek Kraus',
    author_email='zdenek.kraus@gmail.com',
    description='Configurable (structured) data comparator',
    download_url='https://bitbucket.org/zkraus/libracmp/get/libracmp-1.0.1.tar.gz'
)
